@@ -51,8 +51,9 @@ Bool useBGR233 = False;
 Bool forceOwnCmap = False;
 Bool forceTruecolour = False;
 Bool resurface = False;
-Bool reconnect = False;
+Bool reconnect = True;
 int requestedDepth = 0;
+float acceleration = 1.0;
 
 char *geometry = NULL;
 
@@ -109,6 +110,10 @@ void usage()
 	  "              [-noblank]\n"
 	  "              [-lockdelay seconds]\n"
 	  "              [-debug]\n"
+	  "              [-accel multiplier]\n"
+	  "              [-noreconnect]\n"
+	  "              [-tunnel]\n"
+	  "              [-via <host>]\n"
 	  " Known extensions:"
 #ifdef HAVE_XINERAMA
 	  " Xinerama"
@@ -118,6 +123,12 @@ void usage()
 #endif
 #ifdef HAVE_XIDLE
 	  " Xidle"
+#endif
+#ifdef HAVE_XRANDR
+	  " randr"
+#endif
+#ifdef HAVE_XF86DGA
+	  " XFree86-DGA"
 #endif
 	  "\n"
 	  ,programName,programName);
@@ -132,7 +143,6 @@ void processArgs(int argc, char **argv)
   programName = argv[0];
   
   for (i = 1; i < argc; i++) {
-    
     if (strcmp(argv[i],"-display") == 0) {
       
       if (++i >= argc) usage();
@@ -163,6 +173,8 @@ void processArgs(int argc, char **argv)
       wheel_button_up=5;
     } else if (strcmp(argv[i],"-resurface") == 0) {
       resurface=True;
+    } else if (strcmp(argv[i],"-noreconnect") == 0) {
+      resurface=False;
     } else if (strcmp(argv[i],"-shared") == 0) {
       shareDesktop = True;
     } else if (strcmp(argv[i],"-edgewidth") == 0) {
@@ -190,6 +202,9 @@ void processArgs(int argc, char **argv)
     } else if (strcmp(argv[i],"-lockdelay") == 0) {
       if (++i >= argc) usage();
       no_wakeup_delay = atoi(argv[i]);
+    } else if (strcmp(argv[i],"-accel") == 0) {
+      if (++i >= argc) usage();
+      acceleration = atof(argv[i]);
     } else if (strcmp(argv[i],"-scrolllines") == 0) {
       extern int scroll_lines;
       if (++i >= argc) usage();
@@ -215,6 +230,12 @@ void processArgs(int argc, char **argv)
       trimsel = True;
     } else if (strcmp(argv[i],"-noblank") == 0) {
       noblank = True;
+    } else if (strcmp(argv[i],"-tunnel") == 0) {
+      useSSHTunnel=True;
+    } else if (strcmp(argv[i],"-via") == 0) {
+      if (++i >= argc) usage();
+      useSSHGateway = argv[i];
+      useSSHTunnel=True;
     } else if (strcmp(argv[i],"-listen") == 0) {
       if (argumentSpecified) usage();
       
@@ -236,9 +257,7 @@ void processArgs(int argc, char **argv)
 	port += SERVERPORT;
       
     } else {
-      
       usage();
-      
     }
   }
   
