@@ -26,12 +26,31 @@
 #include <sys/types.h>
 #include <unistd.h>
 #include <x2vnc.h>
+#include <math.h>
 #include <X11/X.h>
 #include <X11/Xutil.h>
 #include <X11/Xatom.h>
 #include <X11/cursorfont.h>
 #include <X11/keysym.h>
 
+struct coord
+{
+  int x, y;
+};
+
+struct coord mkcoord(int x, int y)
+{
+  struct coord ret;
+  ret.x=x;
+  ret.y=y;
+  return ret;
+}
+
+void dumpcoord(struct coord *c)
+{
+  fprintf(stderr,"{ %d, %d }",c->x,c->y);
+}
+
 #ifdef HAVE_XIDLE
 #include "xidle.h"
 
@@ -48,6 +67,19 @@ int server_has_mitscreensaver=0;
 #include <X11/extensions/Xinerama.h>
 #endif
 
+#ifdef HAVE_XRANDR
+#include <X11/extensions/Xrandr.h>
+
+int server_has_randr=0;
+int xrandr_event_base;
+int xrandr_error_base;
+#endif
+
+#ifdef HAVE_XF86DGA
+#include <X11/extensions/xf86dga.h>
+int server_has_xf86dga=0;
+#endif
+
 Display *dpy;
 static Window topLevel;
 static int topLevelWidth, topLevelHeight;
@@ -60,10 +92,26 @@ static Bool HandleRootEvent(XEvent *ev);
 static int displayWidth, displayHeight;
 static int x_offset=0, y_offset=0;
 static int grabbed;
-Cursor  grabCursor;
+static int pointer_warp_threshold=5;
+Cursor  grabCursor=0;
+
+float remote_xpos=0.0;
+float remote_ypos=0.0;
+float pointer_speed = 0.0;
+
+struct coord origo1, origo2;
+struct coord current_location;
+struct coord current_speed;
+struct coord * next_origo;
+struct coord * current_origo;
+int origo_separation;
+int motion_events;
 
 #define XROOT(E) ((E).x_root - x_offset)
 #define YROOT(E) ((E).y_root - y_offset)
+#define CENTERX (displayWidth/2+x_offset)
+#define CENTERY (displayHeight/2+y_offset)
+#define ABS(X) ((X) >= 0 ? (X) : -(X))
 
 enum edge_enum edge = EDGE_EAST;
 int edge_width=1;
@@ -261,6 +309,7 @@ Bool CreateXWindow(void)
    }
  }
 #endif
+
   ew=edge_width;
   if(!ew) ew=1;
   topLevelWidth=ew;
@@ -368,8 +417,11 @@ Bool CreateXWindow(void)
     int win_x, win_y;
     unsigned int keys_buttons;
     
-    XSelectInput(dpy, DefaultRootWindow(dpy), PropertyChangeMask |
-		 EnterWindowMask | LeaveWindowMask | KeyPressMask);
+    XSelectInput(dpy, DefaultRootWindow(dpy),
+		 PropertyChangeMask |
+		 EnterWindowMask |
+		 LeaveWindowMask |
+		 KeyPressMask);
     /* Cut buffer happens on screen 0 only. */
     if (DefaultRootWindow(dpy) != RootWindow(dpy, 0)) {
       XSelectInput(dpy, RootWindow(dpy, 0), PropertyChangeMask);
@@ -381,15 +433,25 @@ Bool CreateXWindow(void)
     XSelectInput(dpy, DefaultRootWindow(dpy), PropertyChangeMask);
     mouseOnScreen = 1;
   }
+
+#ifdef HAVE_XRANDR
+ {
+   int major=0, minor=0;
+
+   if(XRRQueryVersion(dpy, &major, &minor) && major >= 1)
+   {
+     XRRQueryExtension(dpy, &xrandr_event_base, &xrandr_error_base);
+     XRRSelectInput(dpy, DefaultRootWindow(dpy), RRScreenChangeNotifyMask);
+     server_has_randr=1;
+   }
+ }
+#endif
   
   nullPixmap = XCreatePixmap(dpy, DefaultRootWindow(dpy), 1, 1, 1);
-#if 0
-  grabCursor = 0;
-#else
-  grabCursor = 
-    XCreatePixmapCursor(dpy, nullPixmap, nullPixmap,
-			&dummyColor, &dummyColor, 0, 0);
-#endif
+  if(!debug)
+    grabCursor = 
+      XCreatePixmapCursor(dpy, nullPixmap, nullPixmap,
+			  &dummyColor, &dummyColor, 0, 0);
   
   if(grabkeysym)
   {
@@ -420,22 +482,62 @@ Bool CreateXWindow(void)
      which keyboard has control */
   SendPointerEvent(restingx,restingy,0);
 
-  
+  pointer_speed = acceleration * pow( (si.framebufferWidth * si.framebufferHeight) / (float)(displayWidth * displayHeight), 0.25 );
+
+#ifdef HAVE_XF86DGA
+ {
+   int major=0, minor=0;
+   if (XF86DGAQueryVersion(dpy, &major, &minor))
+   {
+     int num, den, thr;
+     server_has_xf86dga=1;
+     XGetPointerControl(dpy, &num, &den, &thr);
+     pointer_speed *= num;
+     pointer_speed /= den;
+     pointer_warp_threshold = thr * thr;
+   }
+ }
+#endif
+
+  fprintf(stderr,"%s: pointer multiplier: %f\n",programName,pointer_speed);
+
+
+  /* Compute two identifiable locations, as far as possible from each other */
+  if(displayWidth * 2 > displayHeight)
+  {
+    origo1=mkcoord(displayWidth/3,displayHeight/2);
+    origo2=mkcoord(displayWidth*2/3,displayHeight/2);
+    origo_separation=displayWidth/3;
+  }
+  else if(displayHeight * 2 > displayWidth)
+  {
+    origo1=mkcoord(displayWidth/2,displayHeight/3);
+    origo2=mkcoord(displayWidth/2,displayHeight*2/3);
+    origo_separation=displayHeight/3;
+  }
+  else
+  {
+    int N=(int)( (2*(displayWidth+displayHeight)-sqrt((-3*displayWidth*displayWidth-3*displayHeight*displayHeight+8*displayWidth*displayHeight)) )/7.0 );
+    origo1=mkcoord(N,N);
+    origo2=mkcoord(displayWidth-N,displayHeight-N);
+    origo_separation=N;
+  }
+  origo1.x+=x_offset;
+  origo1.y+=y_offset;
+  origo2.x+=x_offset;
+  origo2.y+=y_offset;
+
+  /*
+  dumpcoord(&origo1);
+  dumpcoord(&origo2);
+  fprintf(stderr,"offset=%d, %d\n",x_offset,y_offset);
+  */
+    
   return True;
 }
 
 
 
-/*
- * ShutdownX.
- */
-
-void
-ShutdownX()
-{
-  XCloseDisplay(dpy);
-}
-
 
 /*
  * check_idle()
@@ -476,6 +578,31 @@ void WiggleMouse(void)
   SendPointerEvent(restingx,restingy,0);
 }
 
+void doWarp(void)
+{
+  if(grabbed)
+  {
+    if(next_origo) return;
+    if(current_origo &&
+       current_location.x == current_origo->x &&
+       current_location.y == current_origo->y)
+      return;
+
+    if(current_origo == &origo1)
+      next_origo=&origo2;
+    else
+      next_origo=&origo1;
+
+    /* fprintf(stderr,"WARP: %d %d\n",next_origo->x, next_origo->y); */
+    XWarpPointer(dpy,None,
+		 DefaultRootWindow(dpy),0,0,0,0,
+		 next_origo->x,
+		 next_origo->y);
+    motion_events=0;
+  }
+}
+
+
 
 /*
  * HandleXEvents.
@@ -494,11 +621,10 @@ Bool HandleXEvents(void)
     if(remote_idle > no_wakeup_delay)
       remote_is_locked=1;
 #ifdef DEBUG
-    fprintf(stderr,"IDLE=%d remote_idle=%d\n",
-	    check_idle(),remote_idle);
+    fprintf(stderr,"IDLE=%d remote_idle=%d noblank=%d remote_is_locked=%d\n", check_idle(),remote_idle,noblank,remote_is_locked);
 #endif
     if(noblank &&
-       remote_idle > 45 &&
+       remote_idle >= 45 &&
        !remote_is_locked &&
        check_idle() < 30000)
     {
@@ -526,6 +652,8 @@ Bool HandleXEvents(void)
     }
   }
 
+  doWarp();
+
   return True;
 }
 
@@ -558,7 +686,7 @@ static int leave_translate(int isedge, int width, int pos)
 ( ((Y)-EDGE(edge==EDGE_NORTH))*(si.framebufferHeight-1)/(displayHeight-1-EDGE(NS)) )
 
 
-static int sendpointerevent(int localx, int localy, int buttonmask)
+static int sendpointerevent(int x, int y, int buttonmask)
 {
   if(mac_mode && (buttonmask & (1<<2)))
   {
@@ -566,7 +694,7 @@ static int sendpointerevent(int localx, int localy, int buttonmask)
     buttonmask |=~ (1<<0);
   }
 
-  return SendPointerEvent(SCALEX(localx), SCALEY(localy), buttonmask);
+  return SendPointerEvent(x, y, buttonmask);
 }
 
 
@@ -602,17 +730,27 @@ static void grabit(int x, int y, int state)
   XGrabKeyboard(dpy, topLevel, True, 
 		GrabModeAsync, GrabModeAsync,
 		CurrentTime);
-  XFlush(dpy);
+
+#ifdef HAVE_XF86DGA
+  if(server_has_xf86dga)
+    XF86DGADirectVideo(dpy, DefaultScreen(dpy), XF86DGADirectMouse);
+#endif
+
+  grabbed=1;
+  next_origo=NULL;
+  current_origo=NULL;
 
   if(x > -1 && y > -1)
   {
-    XWarpPointer(dpy,None, DefaultRootWindow(dpy),0,0,0,0,
-		 x_offset + x,
-		 y_offset + y);
-    sendpointerevent(x, y, (state & 0x1f00) >> 8);
+    doWarp();
+
+    /* Whut? How can this be right? */
+    remote_xpos=SCALEX(x);
+    remote_ypos=SCALEY(y);
+    sendpointerevent(remote_xpos, remote_ypos, (state & 0x1f00) >> 8);
   }
 
-  grabbed=1;
+
   mouseOnScreen = 1;
 
   selection_owner=XGetSelectionOwner(dpy, XA_PRIMARY);
@@ -631,9 +769,16 @@ static void ungrabit(int x, int y, Window warpWindow)
 {
   int i;
 
+#ifdef HAVE_XF86DGA
+  if(server_has_xf86dga)
+    XF86DGADirectVideo(dpy, DefaultScreen(dpy), 0);
+#endif
   SendPointerEvent(restingx,restingy,0);
   if(x > -1 && y > -1 )
+  {
     XWarpPointer(dpy,None, warpWindow, 0,0,0,0, x_offset + x, y_offset + y);
+    XFlush(dpy);
+  }
   XUngrabKeyboard(dpy, CurrentTime);
   XUngrabPointer(dpy, CurrentTime);
   mouseOnScreen = warpWindow == DefaultRootWindow(dpy);
@@ -670,6 +815,44 @@ static void shortsleep(int usec)
 }
 
 
+void dumpMotionEvent(XEvent *ev)
+{
+  fprintf(stderr,"{ %d, %d } -> { %d, %d } = %d, %d\n",
+	  current_location.x,
+	  current_location.y,
+	  XROOT(ev->xmotion),
+	  YROOT(ev->xmotion),
+	  XROOT(ev->xmotion)-current_location.x,
+	  YROOT(ev->xmotion)-current_location.y);
+}
+
+struct coord coord_subtract(struct coord a, struct coord b)
+{
+  return mkcoord(a.x-b.x, a.y-b.y);
+}
+
+struct coord coord_add(struct coord a, struct coord b)
+{
+  return mkcoord(a.x+b.x, a.y+b.y);
+}
+
+int coord_dist_sq(struct coord a, struct coord b)
+{
+  a=coord_subtract(a,b);
+  return a.x*a.x + a.y*a.y;
+}
+
+int coord_dist_from_edge(struct coord a)
+{
+  int n,ret=a.x;
+  if(a.y < ret) ret=a.y;
+  n=displayHeight - a.y;
+  if(n < ret) ret=n;
+  n=displayWidth - a.x;
+  if(n < ret) ret=n;
+  return ret;
+}
+
 /*
  * HandleTopLevelEvent.
  */
@@ -685,7 +868,6 @@ static Bool HandleTopLevelEvent(XEvent *ev)
   /* Atom: requestor asks for a list of supported targets (GRM 24 Oct 2003) */
   static Atom TARGETS;
 
-  
   SET_GRAB_TIMEOUT();
 
   switch (ev->type)
@@ -721,7 +903,7 @@ static Bool HandleTopLevelEvent(XEvent *ev)
            * (GRM 24 Oct 2003)
            */
 	  int ret;
-          int targets[2];
+          Atom targets[2];
 	  if(req->property == None)
 	    req->property = TARGETS;
 
@@ -731,7 +913,7 @@ static Bool HandleTopLevelEvent(XEvent *ev)
 			  req->requestor,
 			  req->property,
 			  XA_ATOM,
-			  8 * sizeof(int),
+			  32,
 			  PropModeReplace,
                           (unsigned char *) targets,
                           2);
@@ -751,7 +933,7 @@ static Bool HandleTopLevelEvent(XEvent *ev)
 	  XChangeProperty(dpy,
 			  req->requestor,
 			  req->property,
-			  XA_STRING,
+			  req->target,
 			  8,
 			  PropModeReplace,
 			  client_selection_text,
@@ -830,7 +1012,8 @@ static Bool HandleTopLevelEvent(XEvent *ev)
 #ifdef DEBUG
 	fprintf(stderr,"Raising window!\n");
 #endif
-	XRaiseWindow(dpy, topLevel);
+	mapwindow();
+	/* XRaiseWindow(dpy, topLevel); */
       }
       return 1;
 
@@ -852,116 +1035,130 @@ static Bool HandleTopLevelEvent(XEvent *ev)
       
       if(grabbed)
       {
-	int i, d;
-        int x, y;
+	int i, d=0;
         Window warpWindow;
-        while (XCheckTypedWindowEvent(dpy, topLevel, MotionNotify, ev))
-	  ;	/* discard all queued motion notify events */
+	struct coord offset=mkcoord(0,0);
+
+	do
+	{
+#ifdef SERVER_HAS_XF86DGA
+	  if(server_has_xf86dga)
+	  {
+	    offset.x+=ev->xmotion.x_root;
+	    offset.y+=ev->xmotion.y_root;
+	  }
+	  else
+#endif
+	  {
+	    struct coord new_location=mkcoord(ev->xmotion.x_root, ev->xmotion.y_root);
+	    if(debug) 
+	    {
+	      dumpMotionEvent(ev);
+	      
+	      if(next_origo)
+		fprintf(stderr," {%d} <? {%d}\n",coord_dist_sq(new_location, *next_origo), coord_dist_sq(new_location, current_location));
+	    }
+	    
+	    motion_events++;
+	    
+	    if(next_origo &&
+	       coord_dist_sq(new_location, *next_origo) < coord_dist_sq(new_location, current_location) ||
+	       coord_dist_from_edge(new_location) < motion_events)
+	    {
+	      current_origo=next_origo;
+	      current_location=*next_origo;
+	      next_origo=NULL;
+	      motion_events=0;
+	    }
+	    
+	    current_speed=coord_subtract(new_location, current_location);
+	    if(current_origo)
+	      offset=coord_add(offset, current_speed);
+	  current_location=new_location;
+	  }
+	}
+        while (XCheckTypedWindowEvent(dpy, topLevel, MotionNotify, ev));
+
+	if(pointer_speed > 1 &&
+	   offset.x*offset.x + offset.y*offset.y < 
+	   pointer_warp_threshold)
+	{
+	  remote_xpos+=offset.x;
+	  remote_ypos+=offset.y;
+	}else{
+	  remote_xpos+=offset.x * pointer_speed;
+	  remote_ypos+=offset.y * pointer_speed;
+	}
 
 #if 0
-	fprintf(stderr,"{ %d, %d } -> { %d, %d }\n",
-		XROOT(ev->xmotion),
-		YROOT(ev->xmotion),
-		SCALEX(XROOT(ev->xmotion)),
-		SCALEY(YROOT(ev->xmotion));
+	fprintf(stderr," ==> {%f, %f} (%d,%d)\n",
+		remote_xpos, remote_ypos,
+		si.framebufferWidth,
+		si.framebufferHeight);
 #endif
-	i=sendpointerevent(XROOT(ev->xmotion),
-			   YROOT(ev->xmotion),
-			   (ev->xmotion.state & 0x1f00) >> 8);
-
-	if(ev->xmotion.state & 0x1f00) return 1;
-
-          /*
-           * Things get complicated for multi-screen X servers,
-           * particularly if the PC screen is "inserted" between two
-           * X screens.
-           * - GRM
-           */
-        /* First, check for normal edges (applies to single screen or
-         * screen edge that doesn't switch screens)
-         */
-        if (ev->xmotion.same_screen)
+	
+	if(!(ev->xmotion.state & 0x1f00))
 	{
-          x = XROOT(ev->xmotion);
-          y = YROOT(ev->xmotion);
-          warpWindow = DefaultRootWindow(dpy);
-	  switch(edge)
-	  {
-	    case EDGE_NORTH: 
-              d=y >= displayHeight-edge_width;
-              y = edge_width;
-              break;
-	    case EDGE_SOUTH:
-              d=y < edge_width;
-              y = displayHeight - edge_width -1;
-              break;
-	    case EDGE_EAST:
-              d=x < edge_width;
-              x = displayWidth -  edge_width -1;
-              break;
-	    case EDGE_WEST:
-              d=x >= displayWidth-edge_width;
-              x = edge_width;
-              break;
+	    warpWindow = DefaultRootWindow(dpy);
+	    switch(edge)
+	    {
+	      case EDGE_NORTH: 
+		d=remote_ypos >= si.framebufferHeight;
+		y = edge_width;  /* FIXME */
+		x = remote_xpos * displayWidth / si.framebufferWidth;
+		break;
+	      case EDGE_SOUTH:
+		d=remote_ypos < 0;
+		y = displayHeight - edge_width -1; /* FIXME */
+		x = remote_xpos * displayWidth / si.framebufferWidth;
+		break;
+	      case EDGE_EAST:
+		d=remote_xpos < 0;
+		x = displayWidth -  edge_width -1;  /* FIXME */
+		y = remote_ypos * displayHeight /si.framebufferHeight ;
+		break;
+	      case EDGE_WEST:
+		d=remote_xpos > si.framebufferWidth;
+		x = edge_width;  /* FIXME */
+		y = remote_ypos * displayHeight / si.framebufferHeight;
+		break;
+	    }
 	  }
-        } else {
-            /*
-             * Different screen. Depending on where the pointer ended up,
-             * we warp to either our default screen or the new screen.
-             * 
-             * If the pointer is "near" the edge we're watching, then the
-             * user moved the pointer off the _opposite_ side of the PC
-             * screen, and the pointer should reappear on the original screen.
-             * 
-             * If not, then the pointer was moved off some other edge - and
-             * we just release the pointer.
-             * 
-             * In either case, the pointer's location - relative to the
-             * screen - is not moved.
-             * 
-             * - GRM
-             */
-          warpWindow = ev->xmotion.root;
-          x = XROOT(ev->xmotion);
-          y = YROOT(ev->xmotion);
-	  switch(edge)
-	  {
-	    case EDGE_NORTH:
-              d=YROOT(ev->xmotion) < displayHeight / 2;
-              break;
-	    case EDGE_SOUTH:
-              d=YROOT(ev->xmotion) > displayHeight / 2;
-              break;
-	    case EDGE_EAST:
-              d=XROOT(ev->xmotion) > displayWidth / 2;
-              break;
-	    case EDGE_WEST:
-              d=XROOT(ev->xmotion) < displayWidth / 2;
-              break;
-	  }
-          if (d) {
-            warpWindow = DefaultRootWindow(dpy);
-          }
-          d = 1;
-              
-        }
-          
+
 	if(d)
 	{
+	  if(x<0) x=0;
+	  if(y<0) y=0;
+	  if(y>=displayHeight) y=displayHeight-1;
+	  if(x>=displayWidth) x=displayWidth-1;
 	  ungrabit(x, y, warpWindow);
 	  return 1;
+	}else{
+	  if(remote_xpos < 0) remote_xpos=0;
+	  if(remote_ypos < 0) remote_ypos=0;
+
+	  if(remote_xpos >= si.framebufferWidth)
+	    remote_xpos=si.framebufferWidth-1;
+
+	  if(remote_ypos >= si.framebufferHeight)
+	    remote_ypos=si.framebufferHeight-1;
+
+	  i=sendpointerevent((int)remote_xpos,
+			     (int)remote_ypos,
+			     (ev->xmotion.state & 0x1f00) >> 8);
 	}
+
       }
       return 1;
-      
-    case ButtonPress:
-    case ButtonRelease:
-      if (emulate_wheel &&
-	  ev->xbutton.button >= 4 &&
-	  ev->xbutton.button <= 5)
-      {
-	int l;
-	if (ev->xbutton.button == wheel_button_up)
+	
+	case ButtonPress:
+	case ButtonRelease:
+	  if (emulate_wheel &&
+	      ev->xbutton.button >= 4 &&
+	      ev->xbutton.button <= 5)
+	  {
+	    int l;
+	    if (ev->xbutton.button == wheel_button_up)
 	  ks = XK_Up;
 	else
 	  ks = XK_Down;
@@ -1011,8 +1208,8 @@ static Bool HandleTopLevelEvent(XEvent *ev)
 		      ~(1 << (ev->xbutton.button - 1)));
       }
       
-      return sendpointerevent(XROOT(ev->xbutton),
-			      YROOT(ev->xbutton),
+      return sendpointerevent(remote_xpos,
+			      remote_ypos,
 			      buttonMask);
 	
     case KeyPress:
@@ -1030,8 +1227,8 @@ static Bool HandleTopLevelEvent(XEvent *ev)
 	 (ev->xkey.state == grabmod ||
 	  (ev->xkey.state & ~ Mod2Mask ) == grabmod ))
       {
-	saved_remote_xpos=XROOT(ev->xkey);
-	saved_remote_ypos=YROOT(ev->xkey);
+	saved_remote_xpos=remote_xpos;
+	saved_remote_ypos=remote_ypos;
 	ungrabit(saved_xpos,saved_ypos,DefaultRootWindow(dpy));
 	return 1;
       }
@@ -1067,7 +1264,6 @@ static Bool HandleTopLevelEvent(XEvent *ev)
       if ((ev->xclient.message_type == wmProtocols) &&
 	  (ev->xclient.data.l[0] == wmDeleteWindow))
 	{
-	    ShutdownX();
 	    exit(0);
 	}
 	break;
@@ -1142,6 +1338,13 @@ static Bool HandleRootEvent(XEvent *ev)
   
   switch (ev->type)
   {
+    default:
+#ifdef HAVE_XRANDR
+      if(ev->type == RRScreenChangeNotify + xrandr_event_base)
+	exit(0);
+#endif
+      break;
+    
     case KeyPress:
     {
       KeySym ks;
